# import unittest
# from lightwood.encoders.audio import AmplitudeTsEncoder


# class TestAmplitudeTsEncoder(unittest.TestCase):
#     def test_encode(self):
#         encoder = AmplitudeTsEncoder()
#         audio_url_arr = ['https://file-examples.com/wp-content/uploads/2017/11/file_example_MP3_700KB.mp3', 'https://file-examples.com/wp-content/uploads/2017/11/file_example_MP3_e1MG.mp3']
#         encoded_audio_arr = encoder.encode(audio_url_arr)
#         for encoded_audio in encoded_audio_arr:
#             pass
